#!/usr/bin/env python

"""
Specific classes for storing position information.

Copyright (C) 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.files import *
from iixr.data import vint, vint_to_array
from array import array

class PositionWriter(FileWriter):

    "Writing position information to files."

    def reset(self):
        self.last_docnum = 0

    def write_positions(self, docnum, positions):

        """
        Write for the document 'docnum' the given 'positions'.
        """

        if docnum < self.last_docnum:
            raise ValueError, "Document number %r is less than previous number %r." % (docnum, self.last_docnum)

        # Make sure that the positions are sorted.

        positions.sort()

        # Write the document number delta.
        # Write the number of positions.

        output = array('B')
        vint_to_array(docnum - self.last_docnum, output)
        vint_to_array(len(positions), output)

        # Write the position deltas.

        last = 0

        for position in positions:
            vint_to_array(position - last, output)
            last = position

        output.tofile(self.f)

        self.last_docnum = docnum

class PositionReader(FileReader):

    "Reading position information within term-specific regions of a file."

    def reset(self):
        self.last_docnum = 0

    def read_positions(self):

        "Read positions, returning a document number and a list of positions."

        # Read the document number delta and add it to the last number.

        self.last_docnum += self.read_number()

        # Read the number of positions.

        npositions = self.read_number()

        # Read the position deltas, adding each previous position to get the
        # appropriate collection of absolute positions.

        i = 0
        last = 0
        positions = []

        while i < npositions:
            last += self.read_number()
            positions.append(last)
            i += 1

        return self.last_docnum, positions

class PositionIndexWriter(FileWriter):

    "Writing position index information to files."

    def reset(self):
        self.last_docnum = 0
        self.last_pos_offset = 0

    def write_positions(self, docnum, pos_offset, count):

        """
        Write the given 'docnum, 'pos_offset' and document 'count' to the
        position index file.
        """

        # Write the document number delta.
        # Write the position file offset delta.
        # Write the document count.

        output = array('B')
        vint_to_array(docnum - self.last_docnum, output)
        vint_to_array(pos_offset - self.last_pos_offset, output)
        vint_to_array(count, output)

        # Actually write the data.

        output.tofile(self.f)

        self.last_pos_offset = pos_offset
        self.last_docnum = docnum

class PositionIndexReader(FileReader):

    "Reading position index information within term-specific regions of a file."

    def reset(self):
        self.last_docnum = 0
        self.last_pos_offset = 0

    def read_positions(self):

        """
        Read a document number, a position file offset for the position index
        file, and the number of documents in a section of that file.
        """

        # Read the document number delta.

        self.last_docnum += self.read_number()

        # Read the offset delta.

        self.last_pos_offset += self.read_number()

        # Read the document count.

        count = self.read_number()

        return self.last_docnum, self.last_pos_offset, count

# Iterators for position-related files.

class IteratorBase:

    "Support for iterating over results."

    def __init__(self, reader):

        "Initialise the iterator using the given 'reader'."

        self.reader = reader
        self.replenish(0) # no iteration initially permitted

    def replenish(self, count):

        "Replenish the iterator with 'count' results."

        self.count = count
        self.read_documents = 0

    def __len__(self):

        "Return the total number of results."

        return self.count

    def sort(self):
        pass # Stored document positions are already sorted.

    def __iter__(self):
        return self

class PositionIterator(IteratorBase):

    "Iterating over document positions."

    def replenish(self, count):
        IteratorBase.replenish(self, count)

        # Fill a cache of positions.

        self.cache = []
        n = 0

        while n < self.count:
            self.cache.append(self.reader.read_positions())
            n += 1

    def seek(self, offset, count):

        """
        Seek to 'offset' in the file, limiting the number of documents available
        for reading to 'count'.
        """

        self.reader.seek(offset)
        self.replenish(count)

    def next(self):

        "Read positions for a single document."

        if self.read_documents < self.count:
            positions = self.cache[self.read_documents]
            self.read_documents += 1
            return positions
        else:
            raise StopIteration

class PositionIndexIterator(IteratorBase):

    "Iterating over document positions."

    def replenish(self, count):
        IteratorBase.replenish(self, count)

        # Fill a cache of offsets.

        self.cache = []
        self.current = 0
        n = 0

        while n < self.count:
            docnum, pos_offset, section_count = t = self.reader.read_positions()
            self.cache.append(t)
            n += section_count

    def seek(self, offset, doc_frequency):

        """
        Seek to 'offset' in the file, limiting the number of documents available
        for reading to 'doc_frequency'.
        """

        self.reader.seek(offset)
        self.replenish(doc_frequency)

    def next(self):

        "Read positions for a single document."

        if self.current < len(self.cache):
            docnum, pos_offset, self.section_count = t = self.cache[self.current]
            self.current += 1
            return t
        else:
            raise StopIteration

class PositionDictionaryWriter:

    "Writing position dictionaries."

    def __init__(self, position_writer, position_index_writer, interval):
        self.position_writer = position_writer
        self.position_index_writer = position_index_writer
        self.interval = interval

    def write_term_positions(self, doc_positions):

        """
        Write all 'doc_positions' - a collection of tuples of the form (document
        number, position list) - to the file.

        Add some records to the index, making dictionary entries.

        Return a tuple containing the offset of the written data, the frequency
        (number of positions), and document frequency (number of documents) for
        the term involved.
        """

        # Reset the writers.

        self.position_writer.reset()
        self.position_index_writer.reset()

        # Remember the first index entry offset.

        index_offset = self.position_index_writer.f.tell()

        # Write the positions.

        frequency = 0
        count = 0

        if doc_positions:

            # Retain the first record offset for a subsequent index entry.

            first_offset = self.position_writer.f.tell()
            first_docnum = None

            doc_positions.sort()

            for docnum, positions in doc_positions:
                if first_docnum is None:
                    first_docnum = docnum

                self.position_writer.write_positions(docnum, positions)

                frequency += len(positions)
                count += 1

                # Every {interval} entries, write an index entry.

                if count % self.interval == 0:

                    self.position_index_writer.write_positions(first_docnum, first_offset, self.interval)

                    first_offset = self.position_writer.f.tell()
                    first_docnum = None

                    # Reset the position writer so that position readers accessing
                    # a section start with the correct document number.

                    self.position_writer.reset()

            # Finish writing an index entry for the remaining documents.

            else:
                if first_docnum is not None:
                    self.position_index_writer.write_positions(first_docnum, first_offset, count % self.interval)

        return index_offset, frequency, count

    def close(self):
        self.position_writer.close()
        self.position_index_writer.close()

class PositionDictionaryReader:

    "Access to position dictionary entries through iterators."

    def __init__(self, position_reader, position_index_reader):
        self.position_reader = position_reader
        self.position_index_reader = position_index_reader

    def read_term_positions(self, offset, doc_frequency):
        iterator = PositionDictionaryIterator(
            PositionIterator(self.position_reader),
            PositionIndexIterator(self.position_index_reader)
            )
        iterator.seek(offset, doc_frequency)
        return iterator

    def close(self):
        self.position_reader.close()
        self.position_index_reader.close()

class PositionDictionaryIterator:

    "Iteration over position dictionary entries."

    def __init__(self, position_iterator, position_index_iterator):
        self.position_iterator = position_iterator
        self.position_index_iterator = position_index_iterator
        self.reset()

    def reset(self):

        # Remember the last values.

        self.found_docnum, self.found_positions = None, None

        # Maintain state for the next index entry, if read.

        self.next_docnum, self.next_pos_offset, self.next_section_count = None, None, None

    def seek(self, offset, doc_frequency):

        """
        Seek to 'offset' in the index file, limiting the number of documents
        available for reading to 'doc_frequency'.
        """

        self.reset()

        # Seek to the appropriate index entry.

        self.position_index_iterator.seek(offset, doc_frequency)

        # Initialise the current index entry and current position file iterator.

        self._next_section()
        self._init_section()

    # Sequence methods.

    def __len__(self):
        return len(self.position_index_iterator)

    def sort(self):
        pass

    # Iterator methods.

    def __iter__(self):
        return self

    def next(self):

        """
        Attempt to get the next document record from the section in the
        positions file.
        """

        # Return any visited but unrequested record.

        if self.found_docnum is not None:
            t = self.found_docnum, self.found_positions
            self.found_docnum, self.found_positions = None, None
            return t

        # Or search for the next record.

        while 1:

            # Either return the next record.

            try:
                return self.position_iterator.next()

            # Or, where a section is finished, get the next section and try again.

            except StopIteration:

                # Although, where a single iterator is in use, the file reader
                # would be positioned appropriately, this is not guaranteed in a
                # multiple iterator situation.

                self._next_section()
                self._init_section()

    def from_document(self, docnum):

        """
        Attempt to navigate to a positions entry for the given 'docnum',
        returning the positions for 'docnum', or None otherwise.
        """

        # Return any unrequested document positions.

        if docnum == self.found_docnum:
            return self.found_positions

        # Read ahead in the index until the next entry refers to a document
        # later than the desired document.

        try:
            if self.next_docnum is None:
                self.next_docnum, self.next_pos_offset, self.next_section_count = self.position_index_iterator.next()

            # Read until the next entry is after the desired document number,
            # or until the end of the results.

            while self.next_docnum <= docnum:
                self._next_read_section()
                if self.docnum < docnum:
                    self.next_docnum, self.next_pos_offset, self.next_section_count = self.position_index_iterator.next()
                else:
                    break

        except StopIteration:
            pass

        # Navigate in the position file to the document.

        self._init_section()

        try:
            while 1:
                found_docnum, found_positions = self.position_iterator.next()

                # Return the desired document positions or None (retaining the
                # positions for the document immediately after).

                if docnum == found_docnum:
                    return found_positions
                elif docnum < found_docnum:
                    self.found_docnum, self.found_positions = found_docnum, found_positions
                    return None

        except StopIteration:
            return None

    # Internal methods.

    def _next_section(self):

        "Attempt to get the next section in the index."

        if self.next_docnum is None:
            self.docnum, self.pos_offset, self.section_count = self.position_index_iterator.next()
        else:
            self._next_read_section()

    def _next_read_section(self):

        """
        Make the next index entry the current one without reading from the
        index.
        """

        self.docnum, self.pos_offset, self.section_count = self.next_docnum, self.next_pos_offset, self.next_section_count
        self.next_docnum, self.next_pos_offset, self.next_section_count = None, None, None

    def _init_section(self):

        "Initialise the iterator for the section in the position file."

        # Seek to the position entry.

        self.position_iterator.seek(self.pos_offset, self.section_count)

# vim: tabstop=4 expandtab shiftwidth=4
